package dm4509.tcp;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;

/**
 * Classe qui
 *  1. établit la connexion d'un client ; 
 *  2. fournit les méthodes d'échange de données sur un canal TCP connecté (par un client ou un serveur).
 * 
 * @author Eric Lallet.
 *
 */
public class TcpSocket implements AutoCloseable {

	/**
	 * un canal TCP connecté (soit par la connexion coté client, soit par l'accept coté serveur)
	 */
	private SocketChannel rwChan = null;
	
	
	/**
	 * Constructeur utilisé par les serveurs pour créer une instance TcpSocket à partir d'une connexion obtenue suite à un accept.
	 * @param rwChan 
	 *              le canal déjà connecté suite à l'accept.
	 */
	public TcpSocket(SocketChannel rwChan) {
		this.rwChan = rwChan;
	}
	
	/**
	 * Constructeur utilisé par le client pour ouvrir une connexion vers un serveur, et créer un instance TcpSocket.
	 * @param serverHost
	 * 	                le nom de la machine où tourne le serveur.
	 * @param serverPort
	 *                  le numéro du port TCP utilisé par le serveur.
	 * @throws IOException
	 *                  toutes les exceptions d'entrées/sorties.
	 */
	public TcpSocket(String serverHost, int serverPort) throws IOException {
		InetSocketAddress rcvAddress = new InetSocketAddress(InetAddress.getByName(serverHost), serverPort);
		this.rwChan = SocketChannel.open(rcvAddress);
	}
	
	/**
	 * Envoie les données contenues dans le buffer sur la connexion TCP.
	 * @param buffer
	 *              le buffer contenant les données à envoyer.
	 * @return
	 *              le nombre d'octets envoyés.
	 * @throws IOException
	 *                  toutes les exceptions d'entrées/sorties.
	 */
	public int sendBuffer(ByteBuffer buffer) throws IOException {	
		return (rwChan.write(buffer));
	}
	
	
	/**
	 * Envoie un objet instancié sérialisable sur la connexion TCP.
	 * @param s
	 *         référence sur l'objet à envoyer.
	 * @return
	 *         le nombre d'octets envoyés.
	 * @throws IOException
	 *          toutes les exceptions d'entrées/sorties.
	 */
	public int sendObject(final Serializable s) throws IOException {
		ByteArrayOutputStream bo = new ByteArrayOutputStream();
		int size;
		
		ObjectOutputStream oo = new ObjectOutputStream(bo);
		oo.writeObject(s);
		oo.close();
		size = bo.size();
		sendSize(size);
		
		ByteBuffer buffer = ByteBuffer.allocate(size);
		buffer.put(bo.toByteArray());
		bo.close();
		buffer.flip();
		
		return sendBuffer(buffer);
	}
	
	/**
	 * Envoie un entier sur la connexion TCP.
	 * @param size
	 *            la valeur de l'entier à envoyer.
	 * @return
	 *            le nombre d'octets envoyés.
	 * @throws IOException
	 *            toutes les exceptions d'entrées/sorties.
	 */
	public int sendSize(int size) throws IOException {
		ByteBuffer buffer = ByteBuffer.allocate(Integer.SIZE/ Byte.SIZE);
		buffer.putInt(size);
		buffer.flip();
		return sendBuffer(buffer);
	}
	
	/**
	 * Reçoit des données sur la connexion TCP jusqu'à la fermeture la connexion entrante ou le remplissage du buffer passé en paramètre.
	 * La méthode remplit le buffer à partir du premier octet de celui-ci, en écrasant éventuellement des données déjà présente.
	 * @param buffer
	 *             le buffer où stocker les données.
	 * @return
	 *             le nombre d'octets placés dans le buffer.
	 * @throws IOException
	 *            toutes les exceptions d'entrées/sorties.
	 */
	public int receiveBuffer(ByteBuffer buffer) throws IOException {
		int lus, total = 0;
		buffer.clear();
		while ((lus = rwChan.read(buffer)) > 0) {
			total += lus;
		}
			
		return total;
		
	}
	
	/**
	 * Reçoit un entier sur la connexion TCP.
	 * @return
	 *         la valeur de l'entier reçu.
	 * @throws IOException
	 *             le nombre d'octets placés dans le buffer.
	 */
	public int receiveSize() throws IOException {
		ByteBuffer buffer = ByteBuffer.allocate(Integer.SIZE/ Byte.SIZE);
		receiveBuffer(buffer);
		buffer.flip();
		return buffer.getInt();
	}
	
	/**
	 * Reçoit un objet sérialisable sur la connexion TCP.
	 * @return
	 *        la référence sur l'objet reçu.
	 * @throws IOException
	 *        toutes les exceptions d'entrées/sorties
	 * @throws ClassNotFoundException
	 *        l'exception levée si l'objet reçu est d'une classe inconnue de notre programme.
	 */
	public Serializable receiveObject() throws IOException, ClassNotFoundException {
		Serializable res = null;
		int size = receiveSize();
		ByteBuffer buffer = ByteBuffer.allocate(size);
		receiveBuffer(buffer);
		buffer.flip();
		
		ByteArrayInputStream bi = new ByteArrayInputStream(buffer.array());
		ObjectInputStream oi = new ObjectInputStream(bi);
		res = (Serializable) oi.readObject();
		oi.close();
		bi.close();
		
		return res;
	}
	
	@Override
	public void close() throws IOException {
		this.rwChan.close();
	}
	
	
	/**
	 * Renvoie tous les octets reçus sur la connexion TCP tant que celle-ci reste ouverte. 
	 * @return
	 *        le nombre total d'octets reçus.
	 * @throws IOException
	 *        toutes les exceptions d'entrées/sorties.
	 */
	public int echo() throws IOException {
		int lus, ecrits, total = 0;
		ByteBuffer buffer = ByteBuffer.allocate(1024);
		
		do {
			buffer.clear();
			lus = rwChan.read(buffer);
			System.out.println("Lus = " + lus);
			if (lus <= 0) {
				break;
			}
			
			total += lus;
			buffer.flip();
			if (lus == 4) {
				int value = buffer.getInt();
				System.out.println("Valeur = " + value);
				buffer.rewind();
			}
			ecrits = rwChan.write(buffer);
			System.out.println("Ecrit  = " + ecrits);
			
		} while (lus > 0);
		
		return total;
	}
	
}
