package tp4509.tcp;


import tp4509.tcp.protcole_action.IndiceColonneAction;
import tp4509.tcp.protcole_action.IndiceLigneAction;
import tp4509.tcp.protcole_action.ValeurGrilleServeurAction;
import tp4509.tcp.protcole_action.ValeurGrilleClientAction;
import tp4509.picross.PicrossApi;
import tp4509.tcp.protcole_action.GagneAction;
/**
 * Liste des types de messages du protocole entre le client et le serveur TCP.
 * Chaque type de message est associé à l'action à réaliser lors de la réception du message.
 * Chaque objet passé en paramètre du constructeur implémente l'interface fonctionnelle
 * ProtocoleAction qui demande l'implémentation d'une méthode:
 *       void realiser(PicrossApi picrossApi, int val1, int val2, int val3);
 * @author Eric Lallet.
 *
 */
public enum Protocole {
	/**
	 * Message serveur -> client qui fournit la taille de la grille au client.
	 */
	TAILLE_GRILLE(new ProtocoleAction() {
		@Override
		public void realiser(final PicrossApi picrossApi, final int val1, final int val2, final int val3) {
			picrossApi.setTailleGrille(val1, val2);
		}
	}),
	/**
	 *  Message serveur -> client qui fournit la taille des grilles d'indices au client.
	 */
	TAILLE_INDICES((PicrossApi picrossApi, int val1, int val2, int val3) -> picrossApi.setTailleIndices(val1, val2)),
	/**
	 * Message serveur -> client qui donne la valeur d'un indice sur les colonnes.
	 */
	INDICE_COLONNE(new IndiceColonneAction()),
	/**
	 * Message serveur -> client qui donne la valeur d'un indice sur les lignes.
	 */
	INDICE_LIGNE(new IndiceLigneAction()),
	/**
	 * Message serveur -> client qui donne la valeur d'une case de la grilel de jeu.
	 */
	VALEUR_GRILLE_SERVEUR(new ValeurGrilleServeurAction()),
	/**
	 * Message serveur -> client qui indique que la partie est gagnée.
	 */
	GAGNE(new GagneAction()),
	/**
	 * Message client -> serveur qui donne la valeur d'une case de la grille de jeu.
	 */
	VALEUR_GRILLE_CLIENT(new ValeurGrilleClientAction());

	/**
	 * Référence vers l'objet qui implémente la méthode 
	 * void realiser(PicrossApi picrossApi, int val1, int val2, int val3)
	 * à appeler lors de la réception du message.
	 */
	private  ProtocoleAction protocoleAction;

	/**
	 * Le constructeur qui associe le type du messager à l'action à réaliser
	 * lors de sa réception.
	 * @param protocoleAction
	 *            Référence vers l'objet qui implémente la méthode à réaliser.
	 */
	Protocole(final ProtocoleAction protocoleAction) {
		this.protocoleAction = protocoleAction;

	}

	/**
	 * getter sur la référence qui implémente la méthode à appeler lors de la réception
	 * du message.
	 * @return
	 *        Référence vers l'objet qui implémente la méthode à réaliser.
	 */
	public ProtocoleAction getProtocoleAction() {
		return protocoleAction;
	}


}
