package tp4509.tcp;

import static dm4509.common.Log.COMM;
import static dm4509.common.Log.LOGGER_NAME_COMM;
import static dm4509.common.Log.LOG_ON;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.StandardSocketOptions;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Level;

import dm4509.common.Log;
import tp4509.picross.FileFormatException;
import tp4509.picross.PicrossApi;

/**
 * Serveur du jeu picross multijoueur.
 * @author Eric Lallet.
 *
 */
public final class ServeurPicross {
	/**
	 * nombre d'arguments attendus par le programme principal.
	 */
	private static final int NBARGS = 2;
	/**
	 * position du port dans argv[].
	 */
	private static  final int PORTARG = 0;
	/**
	 * position du nom du fichier du picross dans  argv[].
	 */
	private static final int FILENAMEARG = 1;
	
	/**
	 * Durée de le pause (en milliseconde) pour le pause de thread.
	 */
	private static final int PAUSE = 2000; 
	
	/**
	 * 
	 */
	private ServeurPicross() {

	}
	/**
	 * Programme principale du serveur.
	 * @param argv
	 *           argv[0] = port TCP du serveur.
	 * @throws IOException
	 *           toutes les exceptions d'entrées/sorties.
	 * @throws FileFormatException 
	 */
	public static void main(final String[] argv) throws IOException, FileFormatException {
		int port = 0;

		ServerSocketChannel listenChannel;


		if (argv.length != NBARGS) {
			System.out.println("usage: ServeurPicross <port> <nomDeFichier>");
			return;
		}
		Log.configureALogger(LOGGER_NAME_COMM, Level.TRACE);
		port = Integer.parseInt(argv[PORTARG]);

		PicrossApi picrossApi = new PicrossApi(argv[FILENAMEARG]);

		listenChannel = createServerSocket(port);
		serverLoop(listenChannel, picrossApi);

	}

	/**
	 * Ouvre un canal serveur non bloquant sur le port TCP numéro port.
	 * @param port
	 *         numéro du port.
	 * @return
	 *         le ServerSocketChannel en attente sur le port numéro port.
	 * @throws IOException
	 *         toutes les exceptions d'entrées/sorties.
	 */
	public static ServerSocketChannel createServerSocket(final int port) throws IOException {
		ServerSocketChannel listenChannel;

		if (LOG_ON && COMM.isInfoEnabled()) {
			COMM.info("[ServeurPicross] Démarage du serveur sur le port " + port);
		}

		listenChannel = ServerSocketChannel.open();
		listenChannel.setOption(StandardSocketOptions.SO_REUSEADDR, true);
		listenChannel.bind(new InetSocketAddress(port));
		listenChannel.configureBlocking(false);

		return listenChannel;

	}

	/**
	 * Boucle du serveur qui gère la connexion de nouveaux clients et l'échange avec tous les clients déjà connectés.
	 * @param listenChannel
	 *        le ServerSocketChannel en attente des nouveaux clients.
	 * @param picrossApi
	 *        référence de la classe d'accès au picross à résoudre.
	 * @throws IOException 
	 *         toutes les exceptions d'entrées/sorties.
	 */
	private static void serverLoop(final ServerSocketChannel listenChannel, final PicrossApi picrossApi) throws IOException {
		Selector selector = Selector.open();
		listenChannel.register(selector, SelectionKey.OP_ACCEPT);
		Map<SelectionKey, ConnexionAsynchrone> connexionsClients = new HashMap<SelectionKey, ConnexionAsynchrone>();
		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[ServeurPicross::ServerLoop] Entrée dans la boucle serveur");
		}


		while (!picrossApi.estGagne()) {
			selector.select();
			Set<SelectionKey> readyKeys = selector.selectedKeys();
			if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
				COMM.trace("[ServeurPicross::ServerLoop] Nb Ready keys:" + readyKeys.size());
			}

			for (SelectionKey curKey: readyKeys) {

				// le cas de l'accept
				if (curKey.isAcceptable()) {
					SocketChannel rwChan = listenChannel.accept();

					// ajout d'un nouveau client
					if (rwChan != null) {
						if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
							COMM.trace("[ServeurPicross::ServerLoop] Nouveau client accepté.");
						}
						rwChan.configureBlocking(false);
						SelectionKey newKey = rwChan.register(selector, SelectionKey.OP_READ);
						ConnexionAsynchrone client = new ConnexionAsynchrone(rwChan);
						connexionsClients.put(newKey, client);
						envoyerPicross(client, picrossApi);
					}
				}

				// le cas des lectures
				if (curKey.isReadable()) {
					ConnexionAsynchrone client = connexionsClients.get(curKey);

					// lecture des données.
					int lus = client.avancerLecture();
					if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
						COMM.trace("[ServeurPicross::ServerLoop] Nouveau message reçu: " + lus);
					}

					// cas de la déconnexion du client.
					if (lus == -1) {
						client.close();
						connexionsClients.remove(curKey);
						curKey.cancel(); // si on ne le fait pas ici, la méthode select() détectera que le canal
						                 // est fermé, et fera cette intruction automatiquement.
						
						if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
							COMM.trace("[ServeurPicross::ServerLoop] Déconnexion d'un client.");
						}
						continue;
					}

					// cas du message complet
					if (client.messageComplet()) {
						if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
							COMM.trace("[ServeurPicross::ServerLoop] Message reçu complet");
						}
						MessageProtocole message = client.getMessage();
						if (message.getType() != Protocole.VALEUR_GRILLE_CLIENT) {
							// Protocole.VALEUR_GRILLE_CLIENT est le seul message attendu par le serveur.
							// on ne stope pas le serveur pour un client HS.
							continue;
						}
						message.getType().getProtocoleAction().realiser(picrossApi, message.getVal1(), message.getVal2(), message.getVal3());
						MessageProtocole mess = new MessageProtocole(Protocole.VALEUR_GRILLE_SERVEUR,  message.getVal1(), message.getVal2(), message.getVal3());
						for (SelectionKey key: connexionsClients.keySet()) {
							if (key != curKey) {
								if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
									COMM.trace("envoi sur " + key);
								}
								connexionsClients.get(key).envoyer(mess);
							}
						}


						// on n'utilisera plus ce message. On prépare la connexion à la réception du message suivant.
						client.clear();
					}
				}

			}

		}

		// la partie est finie on n'accepte plus de nouveau joueur
		listenChannel.close();

		// on signale à tous les clients que la partie est gagnée.
		MessageProtocole mess = new MessageProtocole(Protocole.GAGNE,  0, 0,  0);
		for (SelectionKey key: connexionsClients.keySet()) {
			if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
				COMM.trace("[ServeurPicross::ServerLoop] Partie gagnée" + key);
			}
			connexionsClients.get(key).envoyer(mess);

		}

		// on laisse aux clients le temps  de savoir que la partie est finie
		// avant de fermer la connexion.
		
		try {
			Thread.sleep(PAUSE);
		} catch (InterruptedException e) {

		}


		for (SelectionKey key: connexionsClients.keySet()) {
			connexionsClients.get(key).close();

		}
	}



	/**
	 * Envoi l'état complet du picross (tailles, indices, grille).
	 * @param client
	 *              la connexion reliée au client qui vient de se connecter.
	 * @param picrossApi
	 *              référence de la classe d'accès au picross à résoudre.
	 * @throws IOException
	 *              toutes les exceptions d'entrée/sortie.
	 */
	private static void envoyerPicross(final ConnexionAsynchrone client, final PicrossApi picrossApi) throws IOException {
		int largeur =  picrossApi.getLargeur();
		int hauteur = picrossApi.getHauteur();
		int largeurIndices =  picrossApi.getLargeurIndices();
		int hauteurIndices =  picrossApi.getHauteurIndices();
		MessageProtocole message;

		// envoi des tailles
		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[ServeurPicross::envoyerPicross] Envoi de la taille de la grille:" + largeur + " X " + hauteur + ".");
		}
		message = new MessageProtocole(Protocole.TAILLE_GRILLE, largeur, hauteur, 0);
		client.envoyer(message);

		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[ServeurPicross::envoyerPicross] Envoi de la taille des indices:" + largeurIndices + " X " + hauteurIndices + ".");
		}
		message = new MessageProtocole(Protocole.TAILLE_INDICES, largeurIndices, hauteurIndices, 0);
		client.envoyer(message);



		// envoi des indices sur les colonnes
		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[ServeurPicross::envoyerPicross] Envoie des incides sur les colonnes.");
		}
		for (int col = 0;  col < largeur; col++) {
			for (int lig = 0; lig < hauteurIndices; lig++) {
				int val = picrossApi.getIndicesColonne(col, lig);
				if (val > 0) {
					message = new MessageProtocole(Protocole.INDICE_COLONNE, col, lig, val);
					client.envoyer(message);
				}
			}
		}


		// envoi des indices sur les lignes
		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[ServeurPicross::envoyerPicross] Envoie des incides sur les lignes.");
		}
		for (int col = 0; col < largeurIndices; col++) {
			for (int lig = 0;  lig < hauteur; lig++) {
				int val = picrossApi.getIndicesLigne(col, lig);
				if (val > 0) {
					message = new MessageProtocole(Protocole.INDICE_LIGNE, col, lig, val);
					client.envoyer(message);
				}
			}
		}


		// envoi de la grille de propositions
		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[ServeurPicross::envoyerPicross] Envoie de la grille.");
		}
		for (int lig = 0; lig < hauteur; lig++) {
			for (int col = 0; col < largeur; col++) {
				message = new MessageProtocole(Protocole.VALEUR_GRILLE_SERVEUR, col, lig, picrossApi.getValeur(col, lig));
				client.envoyer(message);
			}
		}


	}

}
