package tp4509.tcp;

import static dm4509.common.Log.COMM;
import static dm4509.common.Log.LOGGER_NAME_COMM;
import static dm4509.common.Log.LOGGER_NAME_GEN;
import static dm4509.common.Log.LOG_ON;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;

import org.apache.log4j.Level;

import dm4509.common.Log;
import tp4509.picross.PicrossApi;
/**
 * Code du thread qui impémente de protocole du client.
 * @author Eric Lallet.
 *
 */
public class ThreadClient implements Runnable {
	/**
	 * Le canal connecté au serveur.
	 */
	private SocketChannel rwChan;
	/**
	 * référence vers la classe qui gère les données du jeu.
	 */
	private  PicrossApi picrossApi;

	/**
	 * Constructeur qui initalise les attributs de la classe.
	 * @param rwChan
	 *        Le canal connecté au serveur.
	 * @param picrossApi
	 *        référence vers la classe qui gère les données du jeu.
	 */
	public ThreadClient(final SocketChannel rwChan, final PicrossApi picrossApi) {
		// picross en réseau.
		Log.configureALogger(LOGGER_NAME_COMM, Level.TRACE);
		Log.configureALogger(LOGGER_NAME_GEN, Level.TRACE);
		this.rwChan = rwChan;
		this.picrossApi = picrossApi;


	}

	@Override
	public void run() {
		MessageProtocole message;

		try { // try pour les ConnexionPerdueException : toujours fatales.

			// *************  Section initialisation du protocole. *************
			try { // try pour les ProtocoleException
				// sur cette section code, les erreurs du protocole sont fatales: le catch stoppe le thread.

				// réception du premier message: doit être TAILLE_GRILLE
				message = recevoirMessage();
				if (message == null) {
					throw new ConnexionPerdueException("La connexion au serveur a été perdue.");
				}
				if (message.getType() != Protocole.TAILLE_GRILLE) {
					throw new ProtocoleException("Erreur de protocole: message attendu : " + Protocole.TAILLE_GRILLE + ", message reçu: "  + message.getType());
				}
				Protocole.TAILLE_GRILLE.getProtocoleAction().realiser(picrossApi, message.getVal1(), message.getVal2(), message.getVal3());

				// réception du deuxième message: doit être TAILLE_INDICES
				message =  recevoirMessage();
				if (message == null) {
					throw new ConnexionPerdueException("La connexion au serveur a été perdue.");
				}
				if (message.getType() != Protocole.TAILLE_INDICES) {
					throw new ProtocoleException("Erreur de protocole: message attendu : " + Protocole.TAILLE_INDICES + ", message reçu: "  + message.getType());
				}
				Protocole.TAILLE_INDICES.getProtocoleAction().realiser(picrossApi, message.getVal1(), message.getVal2(), message.getVal3());

				// les initialisations sont finies, le thread de l'application graphique peut redémarer.
				notifierFinInitialisation();
			} catch (ProtocoleException e) {
				// avant la fin de l'initialisation on ne peut rien faire en cas d'erreur du protocole: fin du jeu.
				System.err.println(e.getMessage());
				return;
			}


			// *************  Section échange des données du protocole. *************
			// la boucle de jeu.
			while (!picrossApi.estGagne()) {
				try { // try pour les ProtocoleException
					// sur cette section de code, on tente de récupérer les erreurs de protocole: on l'affiche, mais on continue.

					// réception de tous les messages sauf TAILLE_INDICES et TAILLE_GRILLE
					message =  recevoirMessage();
					if (message == null) {
						throw new ConnexionPerdueException("La connexion au serveur a été perdue.");
					}
					if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
						COMM.trace("[ThreadClien::run] reception de  " +  message +  ".");
					}
					if ((message.getType() == Protocole.TAILLE_INDICES)
							||	(message.getType() == Protocole.TAILLE_GRILLE)) {
						throw new ProtocoleException("Erreur de protocole: " + message.getType() + " ne doit pas être reçu qu'une fois");
					}

					message.getType().getProtocoleAction().realiser(picrossApi, message.getVal1(), message.getVal2(), message.getVal3());
				} catch (ProtocoleException e) {
					// en cas d'erreur protocolaire à ce stade, on tente de continuer malgrè tout.
					System.err.println(e.getMessage());
				}
			}

		} catch (ConnexionPerdueException e) {
			System.err.println(e.getMessage());
			picrossApi.setConnexionPerdue(true);
		}

	}



	/**
	 * Lit une trame complète du  protocole et retourne le message construit à partir de cette trame.
	 * Méthode à utiliser en mode synchrone (bloquant). Elle se termine si l'un des trois cas suivants arrive:
	 *   1. le buffer est plein (le cas normal);
	 *   2. la connexion est fermée;
	 *   3. une exception se produit.
	 * @return
	 *        le message construit à partir de la trame reçus, ou null si connexion est fermée ou
	 *        si une execption se produit durant la réception.   
	 */
	public  MessageProtocole recevoirMessage() {
		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[TcpSocket::recevoirMessage]");
		}
		// allocation d'un buffer exactement à la bonne taille.
		// buffer plein <==> une trame du protocole est entièrement lue.
		ByteBuffer buffer = ByteBuffer.allocate(MessageProtocole.BUFFERSIZE);
		int lus = 0;
		int total = 0;
		do {
			try {
				lus = rwChan.read(buffer);
				total += lus;
			} catch (IOException e) {
				return null;
			}
		} while (lus > 0);
		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[TcpSocket::recevoirMessage] " + total + " octets reçus sur " + MessageProtocole.BUFFERSIZE + "attendus.");
		}
		if (buffer.hasRemaining()) {
			// la fin de la boucle alors que le buffer n'est pas plein signfifie que la connexion est fermée.
			return null;
		}
		
		// le contructeur de MessageProtocole va lire dans le buffer => il faut un flip() avant de lui donner.
		buffer.flip();
		return new MessageProtocole(buffer);
	}

	/**
	 * Notifie picrossApi que l'initialisation des hauteurs et largeurs est terminée.
	 * Tant que cette initiatilisation n'est pas terminée, picrossApi bloque (wait()) sur les requêtes
	 * qui réclament ces valeurs.
	 */
	public void notifierFinInitialisation() {
		picrossApi.indiquerInitialisationFinie();

	}

}
